/*
 * par: simple paragraph formater
 *
 * (C) 1998 by Kamil Toman (e-mail: toman@horac.ta.jcu.cz)
 *
 * 24th April 1999
 * version: 1.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licences as by published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Publis License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "interfac.h"
#include "hashtab.h"


#define ADD_TO_LIST(b, x) ((x->next) = (b), (b) = (x))
#define HASH(x, l) (((l) * ((*(x) << 7) + (x[l-1]))) % hashsize)

static listTYPE **hashtable;
static unsigned int hashsize;

static INLINE void strlower(unsigned char *str)
{
    int i, end = strlen(str);

    for (i = 0; i <= end; i++)
	str[i] = tolower(str[i]);
}


static void build_hash_table(int count, listTYPE * glist)
{
    int num;
    listTYPE *list;

    hashsize = count * M_PI_2;
    if (!(hashsize % 3))
	hashsize += 2;
    if (!(hashsize % 2))
	hashsize++;

    if ((hashtable = calloc(sizeof(listTYPE *), hashsize)) == NULL)
	err_exit(EXIT_FAILURE, ERRNOMEM);

    while (glist != NULL) {
	list = glist;
	glist = list->next;
	num = HASH(list->item, list->length);
	ADD_TO_LIST(hashtable[num], list);
    }
}


void input_words(int maxlength)
{
    unsigned char *tmp;
    int count = 0;
    listTYPE *list, *glist = NULL;
    FILE *finput;

    ALLOC(tmp, maxlength * sizeof(unsigned char));

    if (inputrc) {
	if ((finput = fopen(inputrc, "r")) == NULL)
	    err_exit(EXIT_FAILURE, ERRNOFIL);

	while (!feof(finput)) {
	    int l = 0;

	    while (l < maxlength) {
		tmp[l] = getc(finput);
		if (tmp[l] == '\n' || tmp[l] == ' ' || tmp[l] == '\t' || feof(finput))
		    break;
		l++;
	    }
	    if (!l)
		continue;
	    tmp[l++] = 0;
	    count++;
	    ALLOC(list, sizeof(listTYPE) + l * sizeof(unsigned char));
	    list->length = l;
	    memcpy(list->item, tmp, l);
	    strlower(list->item);
	    ADD_TO_LIST(glist, list);
	}
    }
    build_hash_table(count, glist);

    free(tmp);
}


int is_in_hash(unsigned char *str, int length)
{
    int num;
    listTYPE *tmp;

    strlower(str);
    num = HASH(str, length);
    tmp = hashtable[num];
    while (tmp != NULL) {
	if (tmp->length == length)
	    if (!strcmp(str, tmp->item))
		return 1;
	tmp = tmp->next;
    }
    return 0;
}


void remove_hash_table(void)
{
    unsigned int i;
    listTYPE *tmp2, *tmp;

    for (i = 0; i < hashsize; i++) {
	tmp = hashtable[i];
	while (tmp != NULL)
	    tmp2 = tmp, tmp = tmp->next, free(tmp2);
    }

}
